from octoprint_smart_pow.lib.data.power_state import API_POWER_STATE_KEY
import pytest
from json import loads
from pathlib import Path
from kasa import SmartPlug
from kasa.tests.newfakes import FakeTransportProtocol
from octoprint_smart_pow.lib.tplink_plug_client import TPLinkPlug
import octoprint.plugin


@pytest.fixture
def api_power_state_off():
    return {API_POWER_STATE_KEY: "Off"}


@pytest.fixture
def api_power_state_on():
    return {API_POWER_STATE_KEY: "On"}


@pytest.fixture(scope="session")
def event_manager():
    octoprint.plugin.plugin_manager(init=True)
    event_manager = octoprint.events.EventManager()
    event_manager.fire(octoprint.events.Events.STARTUP)
    yield event_manager
    event_manager.fire(octoprint.events.Events.SHUTDOWN)


@pytest.fixture
def discovery_data():
    # XXX This file was sourced from https://github.com/python-kasa/python-kasa/blob/a468d520c0856debe171ceeb99aeb3d8ef91ba02/kasa/tests/fixtures/HS100(US)_1.0.json
    # It would be better to dynamically download this fixture file from the repo
    # as part of pytest setup.
    #    The commit should be pulled to the release-version that setup.py declares (and explicitly declare it in setup.py if the dep. isn't already)
    file = Path(__file__).parent / "tests" / "fixtures" / "HS100(US)_1.0.json"
    return loads(file.read_text())


@pytest.fixture
def fake_transport_protocol(discovery_data):
    return FakeTransportProtocol(info=discovery_data)


@pytest.fixture
def backing_smart_device(fake_transport_protocol) -> SmartPlug:
    plug = SmartPlug("127.0.0.1")
    plug.protocol = fake_transport_protocol
    return plug


@pytest.fixture
def tplink_plug_client(backing_smart_device):
    return TPLinkPlug(backing_smart_device)



from octoprint_smart_pow.lib.data.automatic_power_off import (
    AUTOMATIC_POWER_OFF_SCHEDULED_API_KEY,
    ScheduledPowerOffState,
)


def scheduled_power_off_state_to_api_repr(state: ScheduledPowerOffState):
    return {AUTOMATIC_POWER_OFF_SCHEDULED_API_KEY: state.scheduled}


def api_scheduled_power_off_state_to_internal_repr(
    api_state,
) -> ScheduledPowerOffState:
    return ScheduledPowerOffState(
        scheduled=api_state[AUTOMATIC_POWER_OFF_SCHEDULED_API_KEY]
    )



# XXX Rename this to "interval_scheduler.py" and delete the old file after this is confirmed to work for both async and regula routines
# XXX Rename this class to represent running the same routine to run in an interval
from datetime import timedelta
from types import coroutine
from typing import Callable, Coroutine
import sched
import threading
import time
import logging
import asyncio


class AsyncIntervalScheduler:
    """
    Runs a routine at an interval out of the calling thread
    """

    HIGH_PRIORITY = 0

    def __init__(self, routine: Callable, interval: timedelta, logger=logging):
        self.routine = routine
        self.interval_seconds = interval.total_seconds()

        self.t_helper = threading.Thread(target=self.__run)
        self.should_exit = False
        self.logger = logger
        self._has_finished = False

    def __run(self):
        """
        The core logic of the scheduler that also calls the scheduler's routine.
        """
        asyncio.run(self.__loop())

    async def __loop(self):
        while not self.should_exit:
            result = self.routine()
            if asyncio.iscoroutine(result):
                await result

            await asyncio.sleep(self.interval_seconds)
        self._has_finished = True
        self.logger.info("Scheduler exited succesfully!")

    def start(self):
        self.t_helper.start()

    def stop(self):
        self.should_exit = True

    def has_finished(self):
        return self._has_finished

    def running(self):
        """
        Returns whether start() has been called before a stop()
        """
        return self.t_helper.is_alive()


# I used this as a manual test since it was quicker than writing unit-tests. HarHarHar
if __name__ == "__main__":
    squak = lambda: print("Caaw!")
    scheduler = AsyncIntervalScheduler(
        routine=squak, interval=timedelta(seconds=1)
    )
    scheduler.start()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import asyncio
import time
import octoprint.plugin
from octoprint_smart_pow.lib.data.automatic_power_off import (
    AUTOMATIC_POWER_OFF_SCHEDULED_API_KEY,
    AUTOMATIC_POWER_OFF_API_COMMAND,
)
from octoprint_smart_pow.lib.data.power_state import (
    PowerState,
)
from octoprint_smart_pow.lib.data.events import Events
from octoprint_smart_pow.lib import events
from octoprint_smart_pow.lib.event_manager_helpers import (
    fire_automatic_power_off_do_change_event,
    fire_power_state_changed_event,
)
from octoprint_smart_pow.lib.features.automatic_power_off import (
    AutomaticPowerOff,
)
from octoprint_smart_pow.lib.features.power_state_writer import PowerStateWriter
from octoprint_smart_pow.lib.features.power_state_publisher import (
    PowerStatePublisher,
)
from octoprint_smart_pow.lib import discoverer
from octoprint.events import EventManager

import octoprint.plugin
import flask
from octoprint_smart_pow.lib.features.printer_shutdown_predicate import (
    printer_ready_to_shutdown,
)
from octoprint_smart_pow.lib.mappers.automatic_power_off import (
    api_scheduled_power_off_state_to_internal_repr,
    scheduled_power_off_state_to_api_repr,
)

from octoprint_smart_pow.lib.mappers.power_state import (
    power_state_to_api_repr,
    api_power_state_to_internal_repr,
)
from octoprint_smart_pow.lib.data.power_state import (
    API_POWER_STATE_KEY,
    POWER_STATE_DO_CHANGE_API_COMMAND,
    APIPowerState,
)
from octoprint_smart_pow.lib.thread_utils import run_in_thread
from octoprint_smart_pow.lib.tplink_plug_client import TPLinkPlug
import threading
import funcy


class SmartPowPlugin(
    octoprint.plugin.StartupPlugin,
    octoprint.plugin.ShutdownPlugin,
    octoprint.plugin.TemplatePlugin,
    octoprint.plugin.SettingsPlugin,
    octoprint.plugin.AssetPlugin,
    octoprint.plugin.EventHandlerPlugin,
    octoprint.plugin.SimpleApiPlugin,
):
    def __init__(self):
        self.power_publisher = None

    # TODO: documentation for this hook says not to put long running tasks here
    # discoverer.find_tp_link_plug is a long running operation, I should probably
    # do this in a separate thread
    def on_after_startup(self):
        self._logger.info("Starting up Smart Pow Plugin")
        self._logger.info(
            "Discovering TP-Link smart plug device in the home network"
        )
        # TODO Octoprint docs say to not perform long-running or blocking operations in this hook,
        # yet this method can take up to 15 seconds to resolve.
        # reference: https://docs.octoprint.org/en/master/plugins/mixins.html#octoprint.plugin.StartupPlugin.on_after_startup
        self.tp_smart_plug = discoverer.find_tp_link_plug(
            alias=self.__smart_plug_alias_setting(), logger=self._logger
        )
        self.event_manager: EventManager = self._event_bus
        self.power_publisher = PowerStatePublisher(
            event_manager=self.event_manager,
            smart_plug=self.tp_smart_plug,
            logger=self._logger,
        )
        self.power_publisher.start()

        # initialize power controller
        self.power_state_writer = PowerStateWriter(
            plug=self.tp_smart_plug,
            event_manager=self.event_manager,
            logger=self._logger,
        )

        self.automatic_power_off = AutomaticPowerOff(
            self.event_manager,
            funcy.partial(printer_ready_to_shutdown, self._printer),
        )
        self.automatic_power_off.enable()  # TODO: Instead of being hard-coded, we want it controlled by the UI

    def get_settings_defaults(self):
        """
        Defines settings keys and their default values.
        """
        return {
            "tp_link_smart_plug_alias": "3d printer power plug",
        }

    # def get_template_vars(self):
    #     """
    #     Injecting static values into templates

    #     Implemented by TemplatePlugin
    #     """
    #     return dict(power_plug_state=self._settings.get(["power_plug_state"]))

    def register_custom_events(self):
        custom_events = [
            Events.POWER_STATE_CHANGED_EVENT(),
            Events.AUTOMATIC_POWER_OFF_CHANGED_EVENT(),
        ]
        # the order of these operations matter
        Events.set_prefix(f"plugin_smart_pow")
        return custom_events

    def on_event(self, event: str, payload):
        if event == Events.AUTOMATIC_POWER_OFF_CHANGED_EVENT():
            self._logger.info(f"Received event '{event}'")
            self.cond_power_off = payload

    def get_template_configs(self):
        """
        Return a list of configurations for each template
        Each configuration describes properties about the injection of the template.
        """
        return [
            # "type" is the primary key, since by default each type uniquely maps to a specifically named template file
            {"type": "tab", "custom_bindings": True},
        ]

    def __smart_plug_alias_setting(self):
        """Return the alias of the tp-link smart_plug to connect to"""
        return self._settings.get(["tp_link_smart_plug_alias"])

    def on_shutdown(self):
        self.power_publisher.stop()

    def get_assets(self):
        """
        Used by the asset plugin to register custom view models.
        """
        return dict(js=["js/smart_pow.js"])

    # Simple API Plugin Implementation

    def get_api_commands(self):
        return {
            # Each field is the list of all property names this command takes
            # This command is a proxy for the respective event
            POWER_STATE_DO_CHANGE_API_COMMAND: [API_POWER_STATE_KEY],
            AUTOMATIC_POWER_OFF_API_COMMAND: [
                AUTOMATIC_POWER_OFF_SCHEDULED_API_KEY
            ],
        }

    def on_api_command(self, command, data):
        """
        Defining POST route
        """
        import flask  # TODO DO I NEED THIS ?

        # TODO What happens if an exception happens ? Do I need to setup a flask error response
        if command == POWER_STATE_DO_CHANGE_API_COMMAND:
            self.event_manager.fire(
                Events.POWER_STATE_DO_CHANGE_EVENT(), payload=data
            )
        elif command == AUTOMATIC_POWER_OFF_API_COMMAND:
            fire_automatic_power_off_do_change_event(
                api_scheduled_power_off_state_to_internal_repr(data)
            )
        else:
            raise ValueError(f"command {command} is unrecognized")

    def on_api_get(self, request):
        """
        Defining GET route

        Return all relevant data structures since there can only be one GET
        implemented by the SimpleAPIPlugin
        """
        # Wait for dependencies to be defined by startup
        # XXX hacky
        while self.power_publisher is None:
            time.sleep(1)

        api_power_state: APIPowerState = power_state_to_api_repr(
            self.power_publisher.get_state()
        )
        automatic_power_off = scheduled_power_off_state_to_api_repr(
            self.automatic_power_off.get_state()
        )
        return flask.jsonify({**api_power_state, **automatic_power_off})


plugin = SmartPowPlugin()

global __plugin_implementation__
__plugin_implementation__ = SmartPowPlugin()

global __plugin_pythoncompat__
__plugin_pythoncompat__ = ">=2.7,<4"

global __plugin_hooks__
__plugin_hooks__ = {
    "octoprint.events.register_custom_events": plugin.register_custom_events
}



from octoprint_smart_pow.lib.data.automatic_power_off import (
    ScheduledPowerOffState,
)
from octoprint_smart_pow.lib.data.events import Events
from octoprint_smart_pow.lib.data.power_state import PowerState
from octoprint_smart_pow.lib.mappers.automatic_power_off import (
    scheduled_power_off_state_to_api_repr,
)
from octoprint_smart_pow.lib.mappers.power_state import power_state_to_api_repr


def fire_power_state_changed_event(event_manager, power_state: PowerState):
    event_manager.fire(
        event=Events.POWER_STATE_CHANGED_EVENT(),
        payload=power_state_to_api_repr(power_state),
    )


def fire_power_state_do_change_event(event_manager, power_state: PowerState):
    event_manager.fire(
        event=Events.POWER_STATE_DO_CHANGE_EVENT(),
        payload=power_state_to_api_repr(power_state),
    )


def fire_automatic_power_off_do_change_event(
    event_manager, state: ScheduledPowerOffState
):
    event_manager.fire(
        event=Events.AUTOMATIC_POWER_OFF_DO_CHANGE_EVENT(),
        payload=scheduled_power_off_state_to_api_repr(state),
    )


def fire_automatic_power_off_changed_event(
    event_manager, state: ScheduledPowerOffState
):
    event_manager.fire(
        event=Events.AUTOMATIC_POWER_OFF_CHANGED_EVENT(),
        payload=scheduled_power_off_state_to_api_repr(state),
    )



from dataclasses import dataclass
from enum import Enum
from typing import NewType

# Events
class PowerState(Enum):
    ON = "On"
    OFF = "Off"
    UNKNOWN = "Unknown"


# API
POWER_STATE_DO_CHANGE_API_COMMAND = "set_power_state"

API_POWER_STATE_KEY = "power_state"
# Type for a dict object representing PowerState data consumable by APIs
# Required Keys:
#   API_POWER_STATE_KEY: bool
# TODO In python3.8 I can use DictType instead and force required keys
APIPowerState = dict



import asyncio
import time
from datetime import datetime, timedelta
from octoprint_smart_pow.lib.wait_utils import wait_untill, wait_untill_event
import logging
import pytest
import funcy


@pytest.fixture
def mocked_time(mocker):
    mocked_time = mocker.Mock()
    # each call to mocked_time will return the next value in the list
    mocked_time.side_effect = range(10)
    return mocked_time


@pytest.fixture
def mocked_condition(mocker):
    mocked_condition = mocker.Mock()
    mocked_condition.side_effect = [False, False, False, True]
    return mocked_condition


@pytest.mark.asyncio
async def test_wait_untill_passes(mocked_time, mocked_condition):
    await wait_untill(
        condition=mocked_condition,
        poll_period=timedelta(seconds=1),
        timeout=timedelta(seconds=4),
        time=mocked_time,
        sleep=lambda x: True,
    )

    assert len(mocked_condition.call_args_list) == 4


@pytest.mark.asyncio
async def test_wait_untill_fails(mocked_time, mocked_condition):
    with pytest.raises(TimeoutError):
        await wait_untill(
            condition=mocked_condition,
            poll_period=timedelta(seconds=1),
            timeout=timedelta(seconds=3),
            time=mocked_time,
            sleep=lambda x: True,
        )


@pytest.fixture
def publish(event_manager):
    """
    Returns a function that will publish an event
    """

    async def _publish(event, payload=None):
        event_manager.fire(event, payload)

    return _publish


@pytest.mark.asyncio
async def test_wait_untill_event_passes(event_manager, publish):
    await asyncio.gather(
        wait_untill_event(event_manager=event_manager, event="TEST_EVENT"),
        publish(event="TEST_EVENT"),
    )

    await asyncio.gather(
        wait_untill_event(
            event_manager=event_manager, event="TEST_EVENT", payload="hi"
        ),
        publish(event="TEST_EVENT", payload="hi"),
    )


@pytest.mark.asyncio
async def test_wait_untill_event_fails(event_manager, publish):
    """
    Test that wait_untill_event fails when it's supposed to,
    which is when no event with the same name and payload is received

    So we assert that it fails when two events are fired, one
    with a mismatched payload, and the other with a mismatched event
    """
    with pytest.raises(TimeoutError):
        await asyncio.gather(
            wait_untill_event(
                event_manager=event_manager, event="TEST_EVENT", payload="a"
            ),
            # event with mismatched payload
            publish(event="TEST_EVENT", payload="b"),
            # event with mismatched event string
            publish(event="TEST_EVEN", payload="a"),
        )



"""
Find smart plug clients
"""
import asyncio
from funcy import retry, lfilter
from octoprint_smart_pow.lib.tplink_plug_client import TPLinkPlug
from kasa import Discover
import logging


class NoDevicesFoundError(Exception):
    """Indicates that no smart power devices were found on the network"""

    pass


@retry(tries=3, errors=NoDevicesFoundError, timeout=5)
def find_tp_link_plug(alias, logger=logging) -> TPLinkPlug:
    devices = asyncio.run(Discover.discover())
    # XXX can prob use funcy method to select an object from a list that contains a specific property
    def matches_alias(device):
        return device.alias == alias

    matched_devices = lfilter(matches_alias, devices.values())
    if len(matched_devices) == 0:
        logger.warning("No matched devices were found.  Retrying...")
        raise NoDevicesFoundError()
    logger.info("Found TPLink device")
    return TPLinkPlug(matched_devices[0], logger=logging)


if __name__ == "__main__":
    """Manual tests"""
    plug = find_tp_link_plug("3d printer power plug")
    plug.turn_on()
    print(plug.read())



import asyncio
import time
from datetime import timedelta
from octoprint.events import EventManager
from octoprint_smart_pow.lib.event_manager_helpers import (
    fire_power_state_changed_event,
)
from octoprint_smart_pow.lib.smart_plug_client import SmartPlugClient
from octoprint_smart_pow.lib.data.power_state import (
    PowerState,
)
import logging
import kasa
from octoprint_smart_pow.lib.async_interval_scheduler import (
    AsyncIntervalScheduler,
)


class PowerStatePublisher:
    MAX_CONNECTION_FAILED_RETRY_ATTEMPTS = 20
    POLL_INTERVAL_FOR_READ = timedelta(seconds=2)

    """
    Listen to state change events for a a smart power plug, and broadcast them on the EventManager
    """

    def __init__(
        self,
        event_manager: EventManager,
        smart_plug: SmartPlugClient,
        logger=logging,
    ):
        self.logger = logger
        self.event_manager = event_manager
        self.smart_plug = smart_plug

        # An object that will call a routine on an interval
        self.interval_scheduler = AsyncIntervalScheduler(
            routine=self.__publish_if_changed,
            interval=self.POLL_INTERVAL_FOR_READ,
        )
        self.last_updated_state = PowerState.UNKNOWN

    def get_state(self) -> PowerState:
        return self.last_updated_state

    def start(self):
        """
        Start publishing events.
        """
        self.interval_scheduler.start()

    def stop(self):
        """
        Stop publishing events, and cleanup any resources like threads.
        """
        self.interval_scheduler.stop()

    async def __read_current_state(
        self,
        retry_attempts=MAX_CONNECTION_FAILED_RETRY_ATTEMPTS,
        backoff_seconds=5,
    ) -> PowerState:
        """
        Read the state power state, with retries
        """
        attempt = 1
        while True:
            try:
                return await self.smart_plug.read()
            except kasa.exceptions.SmartDeviceException as err:
                if attempt == retry_attempts:
                    raise
                self.logger.warning(
                    "Error when reading Smart Plug state, retry attempt %d/%d after backoff...",
                    attempt,
                    retry_attempts,
                )
                await asyncio.sleep(backoff_seconds)
            finally:
                attempt += 1

    async def __publish_if_changed(self):
        """
        Publishes a "power state changed" event if it has changed since the last time
        this method was called
        """
        current_state = await self.__read_current_state()

        if (
            self.last_updated_state != current_state
            or self.last_updated_state is None
        ):
            self.logger.info(
                "Publisher registered power state changed from %s to %s",
                self.last_updated_state,
                current_state,
            )
            fire_power_state_changed_event(self.event_manager, current_state)
            self.last_updated_state = current_state



import octoprint


def printer_ready_to_shutdown(
    printer: octoprint.printer.PrinterInterface, temperature_threshold=40
):
    """
    Return whether the printer is ready to be turned off
    """
    temp_data = printer.get_current_temperatures()
    if "bed" not in temp_data:
        return False
    # XXX maybe generalize to fetch all tools , which is possible if all tool keys are numbered.
    return (
        temp_data["bed"]["actual"] < temperature_threshold
        and temp_data["tool0"]["actual"] < temperature_threshold
    )



import funcy
from datetime import timedelta
from typing import Callable
import octoprint.printer
import octoprint.events
from octoprint.events import EventManager
from octoprint_smart_pow.lib.async_interval_scheduler import (
    AsyncIntervalScheduler,
)
from octoprint_smart_pow.lib.data.automatic_power_off import (
    ScheduledPowerOffState,
)
from octoprint_smart_pow.lib.data.events import Events
from octoprint_smart_pow.lib.data.power_state import PowerState
from octoprint_smart_pow.lib.event_manager_helpers import (
    fire_automatic_power_off_do_change_event,
    fire_automatic_power_off_changed_event,
    fire_power_state_do_change_event,
)
from octoprint_smart_pow.lib.mappers.automatic_power_off import (
    api_scheduled_power_off_state_to_internal_repr,
)


class AutomaticPowerOff:

    # This is the event passed to event manager
    # It's defined outside of lib/data/events.py because it is specific to this
    # class.
    # TODO consider moving this timer outside of this class in order to support 1:N relationship
    # of timer to client classes.
    #   A good candidate might be adding a class methods in eevnts.py to create timers
    #   and have __init__.py pass the timer event as a constructor to this class.
    TIMEOUT_EVENT = "plugin_smart_pow_timeout"

    def __init__(
        self,
        event_manager: EventManager,
        printer_ready_to_shutdown: Callable[[], bool],
    ):
        self.state = ScheduledPowerOffState(scheduled=False)
        self.printer_ready_to_shutdown = printer_ready_to_shutdown
        self.event_manager = event_manager
        self.timer = None

    def get_state(self) -> ScheduledPowerOffState:
        return self.state

    def __get_print_finished_events(self):
        """
        All events regarding the finishing of a print
        """
        return [
            octoprint.events.Events.PRINT_DONE,
        ]

    def __get_events(self):
        """
        Return all the event's that should be subscribed to
        """
        return [
            self.TIMEOUT_EVENT,
            octoprint.events.Events.PRINT_STARTED,
            octoprint.events.Events.PRINT_DONE,
            Events.AUTOMATIC_POWER_OFF_DO_CHANGE_EVENT(),
        ]

    def __init_timer(self):
        """
        Initializes a new timer.

        @precondition: The current timer should be either not initialized
        or have exited in the past, since this doesn't stop running timers
        """

        def post_timeout_events():
            self.event_manager.fire(
                event=self.TIMEOUT_EVENT,
            )

        if self.timer is None or self.timer.has_finished():
            self.timer = AsyncIntervalScheduler(
                post_timeout_events, timedelta(seconds=5)
            )
        else:
            raise ValueError(
                "Timer object first needs to be stopped before it's garbage collected"
            )

    def enable(self):
        """
        Enable this feature.

        This is needed in order to schedule power offs
        """
        self.__init_timer()
        self.timer.start()
        self.__subscribe(events=self.__get_events(), callback=self.on_event)

    def disable(self):
        """
        Disable this feature.

        A scheduled power-off will automatically be unscheduled
        """
        if not self.enabled:
            raise RuntimeError("Cannot disabled if it hasn't been enabled yet")

        # TODO is self.on_event the same as before ?
        self.__unsubscribe(events=self.__get_events(), callback=self.on_event)
        self.timer.stop()

    @property
    def enabled(self):
        return self.timer is not None and self.timer.running()

    def on_event(self, event: octoprint.events.Events, payload):
        """
        A Finite State Machine (FSM) to implement behavior of smartly turning off
        power
        """
        if event == self.TIMEOUT_EVENT:
            if self.state.scheduled and self.printer_ready_to_shutdown():
                fire_power_state_do_change_event(
                    self.event_manager, power_state=PowerState.OFF
                )
                # After shuttdown power, disable the scheduling automation
                #   or else the system will fight the user the next time they
                #   turn it on!
                fire_automatic_power_off_do_change_event(
                    self.event_manager, ScheduledPowerOffState(scheduled=False)
                )
        elif event == octoprint.events.Events.PRINT_STARTED:
            fire_automatic_power_off_do_change_event(
                self.event_manager, ScheduledPowerOffState(scheduled=False)
            )
        elif event in [octoprint.events.Events.PRINT_DONE]:
            fire_automatic_power_off_do_change_event(
                self.event_manager, ScheduledPowerOffState(scheduled=True)
            )
        elif event == Events.AUTOMATIC_POWER_OFF_DO_CHANGE_EVENT():
            desired_state = api_scheduled_power_off_state_to_internal_repr(
                payload
            )
            if self.state.scheduled != desired_state.scheduled:
                self.state = desired_state
                fire_automatic_power_off_changed_event(
                    self.event_manager, self.state
                )

    def __subscribe(self, events, callback):
        for event in events:
            self.event_manager.subscribe(event, callback)

    def __unsubscribe(self, events, callback):
        for event in events:
            self.event_manager.unsubscribe(event, callback)



class SmartPlugClient:
    """
    Interface for all client's to smart plugs
    """

    def turn_on():
        raise NotImplementedError()

    def turn_off():
        raise NotImplementedError()

    def read():
        raise NotImplementedError()



# def qualify_event_name(event, plugin_name=None):
#     """
#     When we register custom events with Octoprint, Octoprint actually
#     transforms the names to namespace them from other events

#     See this documentation on the transformation
#     https://docs.octoprint.org/en/master/plugins/hooks.html#octoprint-events-register-custom-events
#     """
#     return __quality_event_name(
#         event=event,
#         plugin_name=(plugin_name if not None else __parse_plugin_name())
#     )


# def __quality_event_name(event,plugin_name):
#     return f"plugin_{plugin_name}_{event}"

# def __parse_plugin_name():
#     pass



import asyncio
from kasa.tests.newfakes import FakeTransportProtocol
from octoprint.events import EventManager
from octoprint_smart_pow.lib.data.events import Events
from octoprint_smart_pow.lib.data.power_state import PowerState
import logging
from octoprint_smart_pow.lib.event_manager_helpers import (
    fire_power_state_changed_event,
)

from octoprint_smart_pow.lib.mappers.power_state import (
    api_power_state_to_internal_repr,
)
from octoprint_smart_pow.lib.thread_utils import run_in_thread
from octoprint_smart_pow.lib.tplink_plug_client import TPLinkPlug


class PowerStateWriter:
    """
    Read and write power state using events.
    """

    def __init__(
        self, plug: TPLinkPlug, event_manager: EventManager, logger=logging
    ):
        self.event_manager = event_manager
        self.event_manager.subscribe(
            Events.POWER_STATE_DO_CHANGE_EVENT(), self.change_power_state
        )
        self.plug = plug
        self.logger = logger
        self.current_state: PowerState = PowerState.UNKNOWN

    def get_state(self) -> PowerState:
        return self.current_state

    def change_power_state(self, event, payload):
        desired_state: PowerState = api_power_state_to_internal_repr(payload)
        self.__set_power_state_of_external_device(desired_state)

    # TODO maybe to better indicate in the calling thread, re-raise exceptions
    # from the worker thread ?
    @run_in_thread
    def __set_power_state_of_external_device(self, power_state: PowerState):
        # Clone the plug in order to use this on a new asyncio thread
        plug = self.__clone_plug(self.plug)
        self.logger.info("Issuing turning plug '%s'", power_state.name)
        TIMEOUT_SECS = 10
        if power_state == PowerState.ON:
            asyncio.run(asyncio.wait_for(plug.turn_on(), TIMEOUT_SECS))
        elif power_state == PowerState.OFF:
            asyncio.run(asyncio.wait_for(plug.turn_off(), TIMEOUT_SECS))
        else:
            raise ValueError(f"power_state {power_state} unrecognized")
        fire_power_state_changed_event(self.event_manager, power_state)
        self.logger.info("Finished turning plug '%s'", power_state.name)

    # TODO Generalize to all smart plugs. Offload cloning to the sub-class
    def __clone_plug(self, tp_plug: TPLinkPlug):
        """
        Return a clone of a smart plug.
        """
        cloned_tp_plug = TPLinkPlug(host=tp_plug.plug.host, logger=self.logger)
        # Makes this class testable so the cloned plug
        # uses the same mocked backend, and doesn't actually connect to that host
        if isinstance(tp_plug.plug.protocol, FakeTransportProtocol):
            cloned_tp_plug.plug.protocol = tp_plug.plug.protocol
        return cloned_tp_plug



import pytest
from json import loads
from pathlib import Path
from kasa import SmartPlug
from kasa.tests.newfakes import FakeTransportProtocol
import logging
import asyncio
from octoprint_smart_pow.lib.data.power_state import PowerState
from octoprint_smart_pow.lib.tplink_plug_client import TPLinkPlug


@pytest.mark.asyncio
class TestTPLinkPlugClient:

    # XXX it would be better to implement this precondition that some tests
    # have as a decorator on each test-case.  This is more explicit rather than on
    # reliying on a unit-test
    async def test_device_is_initialized_to_be_off(
        self, backing_smart_device: SmartPlug
    ):
        """
        Make sure the device starts out in the "off" power state
        Some unit-tests depend on this precondition
        """
        await backing_smart_device.update()
        is_on = backing_smart_device.is_on
        assert is_on is False

    async def test_read_power_state(self, tplink_plug_client):
        """Test whether we can read the correct power state"""
        assert await tplink_plug_client.read() is PowerState.OFF

    async def test_set_power_state(self, tplink_plug_client: TPLinkPlug):
        """Test whether we can set the power state"""
        await tplink_plug_client.turn_on()
        assert await tplink_plug_client.read() is PowerState.ON

    async def test_read_power_state_after_external_change(
        self, tplink_plug_client, backing_smart_device: SmartPlug
    ):
        """Test whether we read the correct power state after another system changes it"""
        # this simulates an external system changing the device's power state
        # (versus the client doing it)
        await backing_smart_device.turn_on()
        assert await tplink_plug_client.read() is PowerState.ON



import funcy
from octoprint_smart_pow.lib.data.power_state import (
    API_POWER_STATE_KEY,
    APIPowerState,
)
from octoprint_smart_pow.lib.data.power_state import PowerState


def power_state_to_api_repr(state: PowerState) -> APIPowerState:
    return APIPowerState(**{API_POWER_STATE_KEY: state.value})


def api_power_state_to_internal_repr(api_state: APIPowerState) -> PowerState:
    return PowerState(api_state[API_POWER_STATE_KEY])



import time
from datetime import datetime, timedelta
import funcy
from unittest import mock
import asyncio
from octoprint.events import EventManager


async def wait_untill(
    condition,
    poll_period=timedelta(seconds=1),
    timeout=timedelta(seconds=10),
    condition_name="no_name",
    time=time.time,
    *condition_args,
    **condition_kwargs,
):
    """
    Waits untill the following condition function returns true

    args:
        condition: A zero-arity callable
        poll_period: How often to call the condition
        time: Callable that returns the current time in seconds since epoch
        sleep: Callable that blocks the thread for a certain amount of seconds
        timeout: Total time to wait for the condition
        condition_name: A human friendly name for the condition that will be mentioned in the Timeout error

    throws:
        TimeoutError
    """
    # holds the starting time in seconds since the epoch
    start_time = int(time())
    cond_callable = funcy.partial(condition, *condition_args, **condition_kwargs)
    condition_is_true = cond_callable()
    while (
        int(time()) < start_time + timeout.total_seconds()
        and not condition_is_true
    ):
        await asyncio.sleep(poll_period.total_seconds())
        condition_is_true = cond_callable()

    if not condition_is_true:
        raise TimeoutError(
            f"Waited {timeout} time for condition '{condition_name}' to be True"
        )


async def wait_untill_event(
    event_manager: EventManager,
    event,
    payload=None,
    poll_period=timedelta(seconds=1),
    timeout=timedelta(seconds=10),
):
    try:
        subscriber = mock.Mock()

        def event_was_published():
            return subscriber.call_args == mock.call(event, payload)

        event_manager.subscribe(event=event, callback=subscriber)
        await wait_untill(
            condition=event_was_published,
            poll_period=poll_period,
            timeout=timeout,
            condition_name=f"Event {event} was published",
        )
    finally:
        event_manager.unsubscribe(event=event, callback=subscriber)



# coding=utf-8

########################################################################################################################
### Do not forget to adjust the following variables to your own plugin.

# The plugin's identifier, has to be unique
plugin_identifier = "smart_pow"

# The plugin's python package, should be "octoprint_<plugin identifier>", has to be unique
plugin_package = "octoprint_smart_pow"

plugin_name = "SmartPow"

# The plugin's version. Can be overwritten within OctoPrint's internal data via __plugin_version__ in the plugin module
plugin_version = "0.0.0-rc.4"

# The plugin's description. Can be overwritten within OctoPrint's internal data via __plugin_description__ in the plugin
# module
plugin_description = """An octoprint plugin for turning off and on your printers power supply"""

# The plugin's author. Can be overwritten within OctoPrint's internal data via __plugin_author__ in the plugin module
plugin_author = "Amit Saxena"

# The plugin's author's mail address.
plugin_author_email = "amitsax10@gmail.com"

# The plugin's homepage URL. Can be overwritten within OctoPrint's internal data via __plugin_url__ in the plugin module
plugin_url = "https://github.com/amitSA/OctoPrint-Smart-Pow"

# The plugin's license. Can be overwritten within OctoPrint's internal data via __plugin_license__ in the plugin module
plugin_license = "AGPLv3"

# Any additional requirements besides OctoPrint should be listed here
plugin_requires = ["python-kasa", "funcy", "voluptuous"]

# TODO "voluptuous" is a development dependency for python-kasa, we need it since we
# use python-kasa internals that aren't in the regular dependency list.
# (Is this correct?)
extras_require = {
 'dev': ["pytest-mock", "pytest", "pytest-asyncio", "black", "pylint"]
}

### --------------------------------------------------------------------------------------------------------------------
### More advanced options that you usually shouldn't have to touch follow after this point
### --------------------------------------------------------------------------------------------------------------------

# Additional package data to install for this plugin. The subfolders "templates", "static" and "translations" will
# already be installed automatically if they exist. Note that if you add something here you'll also need to update
# MANIFEST.in to match to ensure that python setup.py sdist produces a source distribution that contains all your
# files. This is sadly due to how python's setup.py works, see also http://stackoverflow.com/a/14159430/2028598
plugin_additional_data = []

# Any additional python packages you need to install with your plugin that are not contained in <plugin_package>.*
plugin_additional_packages = []

# Any python packages within <plugin_package>.* you do NOT want to install with your plugin
plugin_ignored_packages = []

# Additional parameters for the call to setuptools.setup. If your plugin wants to register additional entry points,
# define dependency links or other things like that, this is the place to go. Will be merged recursively with the
# default setup parameters as provided by octoprint_setuptools.create_plugin_setup_parameters using
# octoprint.util.dict_merge.
#
# Example:
#     plugin_requires = ["someDependency==dev"]
#     additional_setup_parameters = {"dependency_links": ["https://github.com/someUser/someRepo/archive/master.zip#egg=someDependency-dev"]}
additional_setup_parameters = {}

########################################################################################################################

from setuptools import setup

try:
    import octoprint_setuptools
except:
    print(
        "Could not import OctoPrint's setuptools, are you sure you are running that under "
        "the same python installation that OctoPrint is installed under?"
    )
    import sys

    sys.exit(-1)

setup_parameters = octoprint_setuptools.create_plugin_setup_parameters(
    identifier=plugin_identifier,
    package=plugin_package,
    name=plugin_name,
    version=plugin_version,
    description=plugin_description,
    author=plugin_author,
    mail=plugin_author_email,
    url=plugin_url,
    license=plugin_license,
    requires=plugin_requires,
    extra_requires=extras_require,
    additional_packages=plugin_additional_packages,
    ignored_packages=plugin_ignored_packages,
    additional_data=plugin_additional_data,
)

if len(additional_setup_parameters):
    from octoprint.util import dict_merge

    setup_parameters = dict_merge(setup_parameters, additional_setup_parameters)

setup(**setup_parameters)



import asyncio
from kasa import SmartPlug
from octoprint_smart_pow.lib.smart_plug_client import SmartPlugClient
from octoprint_smart_pow.lib.data.power_state import PowerState
import logging

# TODO: rename this file to tplink_plug to resemble the new name
class TPLinkPlug(SmartPlugClient):
    """
    Interface for a TPLink smart power plug.
    Implements the "PowerPlugClientInterface"
    """

    pass

    def __init__(
        self, smart_plug: SmartPlug = None, host: str = None, logger=logging
    ):
        """
        Create a plug client either from a kasa SmartPlug object or the host ip
        """
        # If initialized via host, the docs say we need to update() the plug first
        self.needs_initial_async_update = False
        if smart_plug is not None:
            self.plug = smart_plug
        else:
            if host is None:
                raise ValueError(
                    "Host needs to be non-None if smart_plug is None"
                )
            self.plug = SmartPlug(host=host)
            self.needs_initial_async_update = True
        self.logger = logger

    async def turn_on(self):
        if self.needs_initial_async_update:
            await self.plug.update()
            self.needs_initial_async_update = False

        await self.plug.turn_on()

    async def turn_off(self):
        if self.needs_initial_async_update:
            await self.plug.update()
            self.needs_initial_async_update = False

        await self.plug.turn_off()

    async def read(self) -> PowerState:
        """
        Asynchronously read the current power state
        """
        await self.plug.update()
        return PowerState.ON if self.plug.is_on else PowerState.OFF

    def _refresh(self):
        """
        Refresh the internal plug object with it's real state

        Needs to be called before accessing properties for up-to-date data.
        """
        asyncio.run(self.plug.update())
        # self.logger.info("Is event loop open %b",not asyncio.get_event_loop().is_closed())



import funcy


class Events:
    prefix = None

    # TODO: encapsulate this logic into a decorator like this: @event(value="power_changed_event")
    @classmethod
    def POWER_STATE_CHANGED_EVENT(cls):
        return cls.__get_event_name(
            event="power_state_changed_event", prefix=cls.prefix
        )

    @classmethod
    def POWER_STATE_DO_CHANGE_EVENT(cls):
        return cls.__get_event_name(
            event="power_state_do_change_event", prefix=cls.prefix
        )

    # TODO: this should be renamed to something to do with "scheduling"
    # Its confusing what this means right now
    @classmethod
    def AUTOMATIC_POWER_OFF_CHANGED_EVENT(cls):
        return cls.__get_event_name(
            event="conditional_power_off_changed_event", prefix=cls.prefix
        )

    @classmethod
    def AUTOMATIC_POWER_OFF_DO_CHANGE_EVENT(cls):
        return cls.__get_event_name(
            event="conditional_power_off_do_change_event", prefix=cls.prefix
        )

    @classmethod
    def set_prefix(cls, prefix):
        cls.prefix = prefix

    @classmethod
    def __get_event_name(cls, event, prefix=None):
        if prefix is not None:
            return f"{prefix}_{event}"
        return event



from enum import Enum, auto
from dataclasses import dataclass

AUTOMATIC_POWER_OFF_API_COMMAND = "enable_automatic_power_off"
AUTOMATIC_POWER_OFF_SCHEDULED_API_KEY = (
    "automatic_power_off_enabled"  # boolean valued data type
)


@dataclass
class ScheduledPowerOffState:
    scheduled: bool



import pytest
from datetime import timedelta

from octoprint_smart_pow.lib.async_interval_scheduler import (
    AsyncIntervalScheduler,
)
from octoprint_smart_pow.lib.wait_utils import wait_untill


class TestAsyncIntervalScheduler:
    @pytest.mark.asyncio
    async def test_running_routine(self, mocker):
        EXPECTED_ROUTINE_CALLS = 3

        closure = {"calls": 1}

        def routine():
            # We can control how many times this is supposed to be called by the
            # AsyncIntervalScheduler by passing the scheduler via closure,
            # and stopping scheduling after the desired number of calls
            if closure["calls"] == EXPECTED_ROUTINE_CALLS:
                # Based from how the scheduler works, after calling stop()
                # there should be no more calls to routine
                closure["scheduler"].stop()
            closure["calls"] += 1

        # TODO: Once migrated to python3.8, we can also test async routines with mocker.AsyncMock
        mocked_routine = mocker.Mock(wraps=routine)

        scheduler = AsyncIntervalScheduler(
            routine=mocked_routine, interval=timedelta(seconds=0)
        )
        closure["scheduler"] = scheduler

        scheduler.start()
        await wait_untill(
            condition=lambda: scheduler.has_finished(),
            condition_name="scheduler has finished",
        )
        assert len(mocked_routine.call_args_list) == EXPECTED_ROUTINE_CALLS
